import torch
import torch.nn as nn


class RMSNorm(nn.Module):
    def __init__(self, shape=None, dim=(-1, -2), eps=1e-6):
        super().__init__()
        self.dim = dim
        self.shape = shape
        self.scale = nn.Parameter(torch.ones(shape if len(dim) >= 2 else 1))
        self.eps = eps

    def forward(self, x):
        rms = torch.sqrt(x.pow(2).mean(self.dim, keepdims=True) + self.eps)
        scale = self.scale.view_as(rms[0]).unsqueeze(0)
        x_bar = x.div(rms) * scale
        return x_bar


class ConvLayerNorm(nn.Module):
    def __init__(self, num_channels):
        super().__init__()
        self.norm = nn.GroupNorm(num_groups=1, num_channels=num_channels, eps=1e-6)
        self.norm.bias.requires_grad = False
        self.norm.bias.data.mul_(0)

    def forward(self, x):
        return self.norm(x)


import torch
import torch.nn as nn
import torch.nn.functional as F
from x_metaformer.layers.mlp_layers import GeneralizedLinear
from functools import partial


class Attention(nn.Module):
    def __init__(self,
                 channel_loc,
                 dim,
                 head_dim=32,
                 num_heads=4,
                 attn_dropout=0.0,
                 proj_dropout=0.0,
                 scale_value=1.0,
                 num_mem_vecs=0,
                 sparse_topk=0,
                 l2=False,
                 trainable_scale=False,
                 improve_locality=False,
                 **kwargs):
        super().__init__()
        assert channel_loc in [1, -1], 'Token dimension must be 1 or -1.]'
        assert scale_value > 0, 'scale_value must be > 0'
        self.num_heads       = num_heads
        self.trainable_scale = trainable_scale
        self.head_dim        = head_dim
        self.dim             = dim
        self.channel_dim  = channel_loc
        self.sparse_topk  = sparse_topk
        self.num_mem_vecs = num_mem_vecs
        self.improve_locality = improve_locality
        self.l2 = l2
        self.qkv = GeneralizedLinear(in_features=self.dim,
                                     out_features=3*self.head_dim*self.num_heads,
                                     feature_dim=channel_loc,
                                     bias=False
                                     )
        self.proj_out = GeneralizedLinear(in_features=self.head_dim*self.num_heads,
                                          out_features=self.dim,
                                          feature_dim=channel_loc,
                                          bias=False
                                          )
        self.proj_dropout = nn.Dropout(proj_dropout)
        self.attn_dropout = nn.Dropout(attn_dropout)
        self.scale = nn.Parameter(torch.ones(1, num_heads, 1, 1).mul_(scale_value).float(),
                                  requires_grad=trainable_scale)
        if num_mem_vecs > 0:
            self.mem_k = nn.Parameter(torch.randn(num_heads, num_mem_vecs, head_dim))
            self.mem_v = nn.Parameter(torch.randn(num_heads, num_mem_vecs, head_dim))

    def reshape_qkv(self, qkv):
        return qkv.reshape(3, qkv.shape[0], self.num_heads, -1, self.head_dim)

    def reshape_attention(self, attention, x_shape):
        if len(x_shape) > 3:  # initial shape of  B C H W or B H W C
            shape = (-1, *tuple(x_shape[2:])) if self.channel_dim == 1 else (*tuple(x_shape[1:3]), -1)
            return attention.transpose(1, 2).flatten(-2).reshape(x_shape[0], *shape)
        return attention

    def scale_scores(self, scores):
        return scores * (self.head_dim**-0.5 if not self.l2 else 1.0) * self.scale

    def compute_scores(self, q, k):
        if self.l2:
            q, k = (F.normalize(x, dim=-1, p=2) for x in (q, k))
        return q @ k.transpose(-1, -2)

    def cat_mem_vecs(self, k, v, batch_size):
        if self.num_mem_vecs > 0:
            k = torch.cat((k, self.mem_k.repeat(batch_size, 1, 1, 1)), dim=-2)
            v = torch.cat((v, self.mem_v.repeat(batch_size, 1, 1, 1)), dim=-2)
        return k, v

    def sparsify(self, scores):
        if self.sparse_topk and self.sparse_topk < scores.shape[-1]:
            top, _ = scores.topk(self.sparse_topk, dim=-1)
            vk = top[..., -1].unsqueeze(-1).expand_as(scores)
            mask = scores < vk
            val = -torch.finfo(scores.dtype).max
            scores.masked_fill_(mask, val)  # inplace
        return scores

    def adjust_scores(self, scores):
        if self.improve_locality:
            mask = torch.eye(*scores.shape[-2:], device=scores.device).bool()
            mask_value = -torch.finfo(scores.dtype).max
            scores.masked_fill_(mask, mask_value)  # inplace
        return scores

    def forward(self, x):
        qkv = self.reshape_qkv(self.qkv(x))
        q, k, v = qkv.unbind(0)
        k, v = self.cat_mem_vecs(k, v, x.shape[0])
        scores = self.sparsify(self.adjust_scores(self.scale_scores(self.compute_scores(q, k))))
        attention = self.attn_dropout(scores.softmax(-1)) @ v
        out = self.reshape_attention(attention, x.shape)
        return self.proj_dropout(self.proj_out(out))


AttentionConv = partial(Attention, 1)


if __name__ == '__main__':
    for c, f, t in [(64, 16, 8), (128, 22, 44)]:
        test_batch = torch.randn(12, c, f, t)
        print(f'Attention:   {AttentionConv(c, l2=True, scale_value=10, trainable_scale=False)(test_batch).shape}')
    print('TinyTest1 successful')

    for c, f, t in [(64, 16, 8), (128, 22, 44)]:
        test_batch = torch.randn(12, f, t, c)
        print(f'Attention:   {Attention(-1, c, l2=True, scale_value=10, trainable_scale=False)(test_batch).shape}')
    print('TinyTest2 successful')



import torch
import torch.nn as nn
from functools import partial


class ReLUSquared(nn.ReLU):
    """
        ReLUSquared: relu(x) ** 2
        Proposed in https://arxiv.org/abs/2109.08668v2
    """
    def __int__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return super().forward(x)**2


class StarReLU(ReLUSquared):
    """
        StarReLU: s * relu(x) ** 2 + b
        Proposed in https://arxiv.org/abs/2210.13452
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scale = nn.Parameter(torch.ones(1))
        self.bias  = nn.Parameter(torch.zeros(1))

    def forward(self, x):
        return self.scale * super().forward(x) + self.bias


class GatedActFn(nn.Module):
    """
    Gated activation functions
    Proposed in https://arxiv.org/abs/2002.05202
    """
    def __init__(self, act, dim=-1):
        super(GatedActFn, self).__init__()
        self.act = act
        self.dim = dim

    def forward(self, x):
        x, gates = x.chunk(2, dim=self.dim)
        return x * self.act(gates)


GEGLU       = partial(GatedActFn, nn.GELU())
REGLU       = partial(GatedActFn, nn.ReLU())
StarREGLU   = partial(GatedActFn, StarReLU())
RSEGLU      = partial(GatedActFn, ReLUSquared())
SwishGLU    = partial(GatedActFn, nn.SiLU())


class DropPath(nn.Module):
    def __init__(self, drop_prob: float = 0.0):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        if self.drop_prob == 0. or not self.training:
            return x
        keep_prob = 1 - self.drop_prob
        shape = (x.shape[0],) + (1,) * (x.ndim - 1)
        random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
        if keep_prob > 0.0:
            random_tensor.div_(keep_prob)
        return x * random_tensor



import math
from collections.abc import Iterable
from itertools import repeat
import torch
import torch.nn as nn
from x_metaformer.layers.norm_layers import ConvLayerNorm


def _pair(x):
    if isinstance(x, Iterable):
        return tuple(x)
    return tuple(repeat(x, 2))


class Conv2dSame(nn.Conv2d):
    # logic taken from https://github.com/pytorch/pytorch/issues/3867#issuecomment-482711125
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        N, C, H, W = input.shape
        H2 = math.ceil(H / self.stride[0])
        W2 = math.ceil(W / self.stride[1])
        Pr = (H2 - 1) * self.stride[0] + (self.kernel_size[0] - 1) * self.dilation[0] + 1 - H
        Pc = (W2 - 1) * self.stride[1] + (self.kernel_size[0] - 1) * self.dilation[1] + 1 - W
        x_pad = nn.ZeroPad2d((Pr//2, Pr - Pr//2, Pc//2, Pc - Pc//2))(input)
        x_out = self._conv_forward(x_pad, self.weight, self.bias)
        return x_out


class MBConv(nn.Module):
    def __init__(self, dim, expansion=2, kernel_size=3, act=nn.GELU, **kwargs):
        super().__init__()
        med_channels = int(expansion * dim)
        self.conv = nn.Sequential(
            nn.Conv2d(dim, med_channels, (1, 1), padding=0, bias=False),  # pointwise 1x1 in
            act(),
            nn.Conv2d(med_channels, med_channels, kernel_size,            # depthwise
                      padding='same', groups=med_channels, bias=False),
            nn.Conv2d(med_channels, dim, (1, 1), padding=0, bias=False)   # pointwise 1x1 out
        )

    def forward(self, x):
        return self.conv(x)


class ConvDownsampling(Conv2dSame):
    def __init__(self, *args, norm=ConvLayerNorm, pre_norm=False, post_norm=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.norm_in  = norm(self.in_channels) if pre_norm else nn.Identity()
        self.norm_out = norm(self.in_channels) if post_norm else nn.Identity()

    def forward(self, x):
        return self.norm_out(
            super().forward(
                self.norm_in(x)
            )
        )



if __name__ == '__main__':
    import torch
    import torch.nn as nn
    import torch.nn.functional as F


    x = torch.randn(32, 126, 22, 22)
    us = ConvDownsampling(126, 126, kernel_size=3, stride=2)

    print(x.shape, '-->', us(x).shape)


import numpy as np
import torch
import torch.nn as nn
from x_metaformer.layers.conv_layers import MBConv, ConvDownsampling
from x_metaformer.layers.mlp_layers import MLPConv
from x_metaformer.layers.attention_layers import AttentionConv
from x_metaformer.layers.act_layers import StarReLU, ReLUSquared, StarREGLU, DropPath
from x_metaformer.layers.norm_layers import ConvLayerNorm, RMSNorm
from functools import partial
from abc import ABC, abstractmethod
from inspect import signature


class MetaFormerBlock(nn.Module):
    def __init__(self, dim, depth, mixer, drop_probs, norm=ConvLayerNorm,
                 mlp_act=ReLUSquared, mlp_expansion=4, mlp_dropout=0.1, **mixer_kwargs):
        super(MetaFormerBlock, self).__init__()
        self.depth = depth
        self.dim = dim
        self.drop_probs = drop_probs

        self.mixer = nn.ModuleList(
            [nn.Sequential(
                norm(dim),
                mixer(dim, **mixer_kwargs)
            ) for _ in range(depth)]
        )

        self.mlp = nn.ModuleList(
            [nn.Sequential(
                norm(dim),
                MLPConv(dim, mlp_expansion, mlp_dropout, mlp_act)
            ) for _ in range(depth)]
        )

        self.drop_paths      = nn.ModuleList([DropPath(p) for p in drop_probs])
        self.res_scale_mixer = nn.Parameter(torch.ones(depth, dim, 1, 1), requires_grad=True)
        self.res_scale_mlp   = nn.Parameter(torch.ones(depth, dim, 1, 1), requires_grad=True)

    def forward(self, x):
        for i in range(self.depth):
            x = self.res_scale_mixer[i] * x + self.drop_paths[i](self.mixer[i](x))
            x = self.res_scale_mlp[i]   * x + self.drop_paths[i](self.mlp[i](x))
        return x


def posemb_sincos_2d(patches, temperature = 10000):
    b, dim, h, w, device, dtype = *patches.shape, patches.device, patches.dtype

    y, x = torch.meshgrid(torch.arange(h, device = device), torch.arange(w, device = device), indexing = 'ij')
    assert (dim % 4) == 0, 'feature dimension must be multiple of 4 for sincos emb'
    omega = torch.arange(dim // 4, device = device) / (dim // 4 - 1)
    omega = 1. / (temperature ** omega)

    y = y.flatten()[:, None] * omega[None, :]
    x = x.flatten()[:, None] * omega[None, :]
    pe = torch.cat((x.sin(), x.cos(), y.sin(), y.cos()), dim = 1)
    return pe.type(dtype).view(1, dim, h, w)


def _init_weights(m):
    if isinstance(m, nn.Conv2d):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0)
        if m.kernel_size != (1, 1):
            nn.init.kaiming_normal_(m.weight)
        else:
            nn.init.trunc_normal_(m.weight, std=.02)


class MetaFormerABC(nn.Module, ABC):
    def __init__(self,
                 in_channels,
                 mixers,
                 depths=(3, 3, 9, 3),
                 dims=(64, 128, 256, 320),
                 norm='ln',
                 use_starreglu=False,
                 **kwargs
                 ):
        super().__init__()
        assert 2 <= len(dims) == len(depths) >= 2
        self.in_channels = in_channels
        self.mixers = mixers
        self.dims = dims
        self.depths = depths
        self.use_starreglu = use_starreglu
        self.act = StarReLU if not use_starreglu else StarREGLU

        self.norm_inner, self.norm_out = self.get_norm(norm)

        self.pooling: nn.ModuleList
        self.blocks: nn.ModuleList

        self.out_dim = dims[-1]
        self.norm_out = self.norm_out(self.out_dim)

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass

    def get_norm(self, mode: str):
        if mode == 'ln':
            return ConvLayerNorm, nn.LayerNorm
        elif mode == 'rms':
            return partial(RMSNorm, dim=(-2, -1)), partial(RMSNorm, dim=(1, ))
        elif mode == 'bn':
            return nn.BatchNorm2d, nn.BatchNorm1d
        else:
            raise NotImplemented('Norm must be "ln", "rms" or "bn"')


class MetaFormer(MetaFormerABC):
    def __init__(self,
                 *args,
                 init_kernel_size=3,
                 init_stride=2,
                 drop_path_rate=0.3,
                 use_pos_emb=True,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.init_kernel_size = init_kernel_size
        self.init_stride = init_stride
        self.drop_path_rate = drop_path_rate
        self.use_pos_emb = use_pos_emb

        new_args = set(kwargs.keys()) - set(signature(super().__init__).parameters.keys())
        new_args = {k: v for k, v in kwargs.items() if k in new_args}
        self.mixer_kwargs = new_args  # every kwarg that is not already in the args of MetaFormerABC

        init_downsampling = ConvDownsampling(self.in_channels, self.dims[0],
                                             self.init_kernel_size, self.init_stride,
                                             norm=self.norm_inner, pre_norm=False, post_norm=False
                                             )

        dp_rates = [x.item() for x in torch.linspace(0, self.drop_path_rate, sum(self.depths))]
        dix = np.cumsum((0, *self.depths))

        self.pooling = nn.ModuleList([init_downsampling] + [
            ConvDownsampling(self.dims[i], self.dims[i+1],
                             kernel_size=3, stride=2,
                             pre_norm=True, norm=self.norm_inner) for i in range(len(self.dims)-1)
        ])

        self.blocks = nn.ModuleList([
            MetaFormerBlock(self.dims[i], depth=self.depths[i], mixer=self.mixers[i], norm=self.norm_inner,
                            mlp_act=self.act, act=StarReLU, drop_probs=dp_rates[dix[i]: dix[i+1]], **self.mixer_kwargs)
            for i in range(len(self.depths))
        ])

        self.apply(_init_weights)

    def pool(self, x):
        return x.mean([-2, -1])

    def forward(self, x, return_embeddings=False):
        for i in range(len(self.blocks)):
            x = self.pooling[i](x)
            if i == 0 and self.use_pos_emb:
                x = x + posemb_sincos_2d(x)
            x = self.blocks[i](x)
        pooled = self.norm_out(self.pool(x))
        return pooled if not return_embeddings else (pooled, x)


class ConvFormer(MetaFormer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args,
                         mixers=(MBConv, MBConv, MBConv, MBConv),
                         **kwargs)


class CAFormer(MetaFormer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args,
                         mixers=(MBConv, MBConv, AttentionConv, AttentionConv),
                         **kwargs)


if __name__ == '__main__':
    x = torch.randn(64, 3, 64, 64)
    encoder = CAFormer(3, norm='ln', depths=(2, 2, 4, 2),
                       dims=(16, 32, 64, 128), init_kernel_size=3, init_stride=2)
    codes = encoder(x, return_embeddings=True)[-1]
    print('CODES', codes.shape)


import torch.nn as nn
from x_metaformer.layers.act_layers import GatedActFn


class GeneralizedLinear(nn.Module):
    def __init__(self, in_features, out_features, bias=True, feature_dim=-1):
        super().__init__()
        assert feature_dim in [1, -1], 'Only first or last dimension supported'
        self.linear = nn.Linear(in_features, out_features, bias=bias) if feature_dim == -1 else \
                      nn.Conv2d(in_features, out_features, (1, 1), padding=0, bias=bias)

    def forward(self, x):
        return self.linear(x)


class MLP(nn.Module):
    def __init__(self, dim, expansion=4, p_dropout=0.1, act=nn.GELU, bias=False, feature_dim=-1):
        super().__init__()
        med_channels = int(expansion * dim)
        is_gated = isinstance(act(), GatedActFn)
        self.mlp = nn.Sequential(
            GeneralizedLinear(dim,
                              med_channels if not is_gated else med_channels*2,
                              bias,
                              feature_dim),
            act(feature_dim if is_gated else  None),
            nn.Dropout(p_dropout),
            GeneralizedLinear(med_channels, dim, bias, feature_dim),
            nn.Dropout(p_dropout)
        )

    def forward(self, x):
        return self.mlp(x)


class MLPConv(MLP):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, feature_dim=1, **kwargs)


class FeatureRefiner(nn.Module):
    # https://arxiv.org/pdf/2210.05657.pdf
    def __init__(self, in_features, hidden_dim=64, act=nn.ReLU):
        super(FeatureRefiner, self).__init__()
        self.in_features = in_features
        self.hidde_dim = hidden_dim
        self.refiner = nn.Sequential(
            nn.Linear(in_features, hidden_dim),
            nn.LayerNorm(hidden_dim),
            nn.Linear(hidden_dim, hidden_dim),
            act(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.LayerNorm(hidden_dim)
        )

    def forward(self, x):
        return self.refiner(x)



from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(name='x-Metaformer',
      version='0.1.6',
      description='A PyTorch implementation of "MetaFormer Baselines" with optional extensions.',
      author='Robert Müller',
      author_email='robert.mueller1990@googlemail.com',
      url='https://github.com/romue404/x-metaformer',
      license='MIT',
      keywords=[
            'artificial intelligence',
            'pytorch',
            'metaformer',
            'transformer',
            'attention',
            'convolutions'
      ],
      classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.6',
      ],
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages=find_packages(exclude=['examples']),
      install_requires=['torch>=1.6', 'numpy']
      )
